/* ------------------------------------------------------------------------ */
/* Virtual Memory Library (VML) Version 1.3                                 */
/*                                                                          */
/* (c) Copyright 1998-1999 by De Trans Software (D.T.S.)                    */
/*                                                                          */
/* For questions or suggestions contact us at:                              */
/*     De Trans Software (D.T.S.)                                           */
/*     Javier Gutirrez Chamorro                                            */
/*     Av. Meridiana 336, 7 C                                               */
/*     08027. Barcelona. [SPAIN]                                            */
/*                                                                          */
/*     dts@geocities.com                                                    */
/*     http://www.geocities.com/SiliconValley/4505                          */
/* ------------------------------------------------------------------------ */
/*   This library is intended for users of the Borland C compilers family   */
/* (Turbo C, Turbo C++, Borland C, and Borland C++), althought it is easy   */
/* to port to other compilers.                                              */
/*                                                                          */
/*   The library itself performs a very simple implementation of virtual    */
/* memory (at the same time with a good performance in speed terms).        */
/*                                                                          */
/*   The API implemented is very similar to the XMS API, it simplifies the  */
/* usage of one or more of this memory managements; and only includes the   */
/* basic "starter kit".                                                     */
/* register by sending 1000 Ptas or 10 US$ to:                              */
/*                                                                          */
/*   You can use, copy and distribute this code freely, but please include  */
/* this header, or at least our copyright or thank to us.                   */
/* ------------------------------------------------------------------------ */

/* -------------------------- Header files -------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <string.h>
#include <dos.h>
#include <bios.h>

/* ------------------------ Global variables ------------------------------ */
char virmsg[]=  "\r\n"
                "Virtual Memory Library (VML) Version 1.3\r\n"
								"(c) Copyright 1998-1999 by De Trans Software (D.T.S.)\r\n"
								"De Trans Software (D.T.S.)\r\n"
								"Javier Gutirrez Chamorro\r\n"
								"Av. Meridiana 336, 7 C\r\n"
								"08027. Barcelona. [SPAIN]\r\n"
								"dts@geocities.com\r\n"
								"http://www.geocities.com/SiliconValley/4505\r\n";

unsigned int virhandles=1;
char virdir[80]="C:\\";

/* ----------------------- Function prototypes ---------------------------- */
unsigned int pascal virinit (void);
void pascal virinfo (char *);
unsigned long int pascal virtotal (void);
unsigned long int pascal viravail (void);
unsigned int pascal viralloc (unsigned long int);
unsigned int pascal virfree (unsigned int);
unsigned int pascal vir2con (unsigned int, void *, unsigned int, signed long);
unsigned int pascal con2vir (void *, unsigned int, unsigned int, signed long);

/* ------------------------- Run the demo --------------------------------- */
void main (void)
{
  #define MAXBLOCK 1000L
	#define MAXFRAMES 25
	register unsigned int i, handle;
	register unsigned char y;
	unsigned int handles[MAXFRAMES];
	union REGS r;
	char buffer[1024];


  puts("Now you will see how to allocate, free, and get information of your");
  puts("virtual memory.");
  puts("Press any key...");
	bioskey(0);
	/* Initialize virtual memory */
	if (virinit()!=0)
	{
		puts("Error initializing virtual memory");
		return;
	}
	/* Get info about virtual memory */
	virinfo(buffer);
	printf("Virtual memory path is %s\n", buffer);

	/* Get total virtual memory in Kb */
	printf("Total virtual memory is %ldK\n", virtotal()>>10);

	/* Get available virtual memory in Kb */
	printf("Available virtual memory is %ldK\n", viravail()>>10);

	/* Allocate MAXBLOCK K */
	printf("Allocating %uK...\n", MAXBLOCK);
	handle=viralloc(MAXBLOCK<<10);
	if (handle==0)
	{
		puts("Error allocating virtual memory");
    return;
	}

	/* Fill the allocated memory with 0s, using 1024 bytes (1 K) groups */
	memset(buffer, 0, 1024);
	for (i=0; i<MAXBLOCK; i++)
	{
		printf("Filling block %u of %u\r", i, MAXBLOCK);
		con2vir(buffer, handle, 1000, i*1000L);
	}
	printf("Filled %u blocks (of 1024 bytes each one)\n", MAXBLOCK);

	/* Free virtual memory */
	puts("Freeing virtual memory...");
	virfree(handle);
  puts("Press any key...");
	bioskey(0);


  puts("Now you will see colorized lines in your screen. First of all they are");
  puts("calculated, drawed to the screen, and then saved to the virtual memory");
  puts("Then they are showed forwards and backwards until a key is pressed.");
  puts("Press any key...");
	bioskey(0);
	/* Initialize virtual memory */
	if (virinit()!=0)
	{
		puts("Error initializing virtual memory");
    return;
	}
	if (viravail()<MAXFRAMES*100000L)
	{
		printf("At least %ld bytes of swap space is required", MAXFRAMES*100000L);
    return;
	}
  /* Set the 320x200x256 graphics mode */
	r.x.ax=0x13;
	int86(0x10, &r, &r);
  /* Calculate the lines and save them in virtual memory */
	for (i=0; i<MAXFRAMES; i++)
	{
		handles[i]=viralloc(64000U);
		if (handles[i]==0)
		{
			puts("Error allocating memory");
      return;
		}
		for (y=0; y<200; y++)
			memset(MK_FP(0xA000, y*320), y+i, 320);
		printf("\rFrame %02u\r", i+1);
		con2vir(MK_FP(0xA000, 0x0000), handles[i], 64000U, 0);
	}
  /* Draw each frame until a key is pressed */
  do
  {
    /* Backwards */
	  for (i=0; i<MAXFRAMES; i++)
  	{
  		if (bioskey(1))
	    {
    	  i=0xF000;
      	break;
	    }
  		vir2con(handles[i], MK_FP(0xA000, 0x0000), 64000U, 0);
      delay(100);
	  }
    /* Backwards */
	  for (i=0; i<MAXFRAMES; i++)
  	{
  		if (bioskey(1))
	    {
    	  i=0xF000;
      	break;
	    }
  		vir2con(handles[MAXFRAMES-i], MK_FP(0xA000, 0x0000), 64000U, 0);
      delay(100);
	  }
  }
  while (i<0xF000);
	/* Free all the memory */
	for (i=0; i<MAXFRAMES; i++)
	{
		virfree(handles[i]);
	}
  /* Back to text */
	r.x.ax=0x3;
	int86(0x10, &r, &r);
  puts("Press any key...");
	bioskey(0);
  printf("Now the license (legal stuff):\n%s", virmsg);
  puts("Press any key...");
	bioskey(0);
}

/* ---------------------------- virinit ----------------------------------- */
/* Function:    Initializes the virtual memory library                      */
/* Parameters:  None                                                        */
/* Return:      0 if initialization successful                              */
/* ------------------------------------------------------------------------ */
unsigned int pascal virinit (void)
{
	char *p;

	if ((p=getenv("TEMP"))!=0)
		strcpy(virdir, p);
	else if ((p=getenv("TMP"))!=0)
		strcpy(virdir, p);
	else
		sprintf(virdir, "%c:\\", 'A' + getdisk());

	if (virdir[strlen(virdir)-1]!='\\')
		strcat(virdir, "\\");
	strupr(virdir);
	return(0);
}

/* ---------------------------- virinfo ----------------------------------- */
/* Function:    Get information about the virtual memory                    */
/* Parameters:  info: A buffer where to fit the information (path of the    */
/*              virtual memory files).                                      */
/* Return:      None                                                        */
/* ------------------------------------------------------------------------ */
void pascal virinfo (char *info)
{
	strcpy(info, virdir);
}

/* ---------------------------- virtotal ---------------------------------- */
/* Function:    Get total virtual memory                                    */
/* Parameters:  None                                                        */
/* Return:      Total bytes of virtual memory, or 0 if error or not         */
/*              available.                                                  */
/* ------------------------------------------------------------------------ */
unsigned long int pascal virtotal (void)
{
	register int drive;
	long total;
	struct dfree free;

	drive=virdir[0]-'A';
	getdfree(drive+1, &free);
	if (free.df_sclus==0xFFFF)
		return(0);

	total=(long) free.df_total *
				(long) free.df_bsec *
				(long) free.df_sclus;

	return(total);
}

/* ---------------------------- viravail ---------------------------------- */
/* Function:    Get available virtual memory                                */
/* Parameters:  None                                                        */
/* Return:      Available bytes of virtual memory, or 0 if error or not     */
/*              available.                                                  */
/* ------------------------------------------------------------------------ */
unsigned long int pascal viravail (void)
{
	register int drive;
	long avail;
	struct dfree free;

	drive=virdir[0]-'A';
	getdfree(drive+1, &free);
	if (free.df_sclus==0xFFFF)
		return(0);

	avail=(long) free.df_avail *
				(long) free.df_bsec *
				(long) free.df_sclus;

	return(avail);
}

/* ---------------------------- viralloc ---------------------------------- */
/* Function:    Allocate a virtual memory block                             */
/* Parameters:  size: Size in bytes to allocate                             */
/* Return:      A valid virtual memory handle, or 0 if error                */
/* ------------------------------------------------------------------------ */
unsigned int pascal viralloc (unsigned long int size)
{
	register signed int h;
	unsigned long l;
	char s1[80];

	sprintf(s1, "%sMMS%05u.TMP", virdir, virhandles);
	h=_creat(s1, FA_ARCH);
	if (h<0)
		return(0);
	l=0;
	do
	{
		if ((size-l)>65534U)
			_write(h, 0, 65534U);
		else
			_write(h, 0, size-l);
		l+=65534U;
	}
	while (l<size);
	_close(h);
	virhandles++;
	return(virhandles-1);
}

/* ---------------------------- viralloc ---------------------------------- */
/* Function:    Free a previously allocated a virtual memory block          */
/* Parameters:  handle: Handle of the virtual memory block to free          */
/* Return:      0 if success                                                */
/* ------------------------------------------------------------------------ */
unsigned int pascal virfree (unsigned int handle)
{
	char s1[80];

	sprintf(s1, "%sMMS%05u.TMP", virdir, handle);
	if (remove(s1))
  	return(1);
  else
		return(0);
}

/* ----------------------------- vir2con ---------------------------------- */
/* Function:    Transferres a block from the virtual memory to a            */
/*              conventional memory buffer                                  */
/* Parameters:  handle: Handle of the virtual memory transfer               */
/*              buffer: Buffer where to fit the data                        */
/*              size:   Number of bytes to move                             */
/*              offset: Position of the virtual memory handle to get        */
/* Return:      Number of bytes transferred (0 if error)                    */
/* ------------------------------------------------------------------------ */
unsigned int pascal vir2con (unsigned int handle, void *buffer, unsigned int size, signed long offset)
{
	register unsigned int i;
	register signed int h;
	char s1[80];

	sprintf(s1, "%sMMS%05u.TMP", virdir, handle);
	h=_open(s1, O_RDONLY | O_BINARY);
	if (h<0)
		return(0);
	if (lseek(h, offset, SEEK_SET)<0)
		return(0);
	offset=tell(h);
	i=_read(h, buffer, size);
	_close(h);
  if (i==0xFFFF)
  	return(0);
  else
		return(i);
}

/* ----------------------------- vir2con ---------------------------------- */
/* Function:    Transferres a block from conventional memory buffer to      */
/*              a virtual memory                                            */
/* Parameters:  buffer: Buffer where the data is                            */
/*              handle: Handle of the virtual memory transfer               */
/*              size:   Number of bytes to move                             */
/*              offset: Position of the virtual memory handle to put        */
/* Return:      Number of bytes transferred (0 if error)                    */
/* ------------------------------------------------------------------------ */
unsigned int pascal con2vir (void *buffer, unsigned int handle, unsigned int size, signed long offset)
{
	register unsigned int i;
	register signed int h;
	char s1[80];

	sprintf(s1, "%sMMS%05u.TMP", virdir, handle);
	h=_open(s1, O_WRONLY | O_BINARY);
	if (h<0)
		return(0);
	if (lseek(h, offset, SEEK_SET)<0)
		return(0);
	i=_write(h, buffer, size);
	_close(h);
    if (i==0xFFFF)
  	return(0);
  else
		return(i);
}

